# Copyright (c) 2021 Damon Lynch
# SPDX - License - Identifier: MIT

import platform

from ..constants import Platform
from . import linux

system = platform.system()
is_wsl = False
is_wsl1 = False
is_wsl2 = False
if system == "Windows":
    current_platform = Platform.windows
elif system == "Linux":
    current_platform = Platform.linux
    if linux.detect_wsl():
        is_wsl = True
        if linux.wsl_version() == linux.LinuxDesktop.wsl2:
            is_wsl2 = True
        else:
            is_wsl1 = True
elif system == "Darwin":
    current_platform = Platform.macos
else:
    current_platform = None
    raise NotImplementedError



# Copyright (c) 2016-2021 Damon Lynch
# SPDX - License - Identifier: MIT


import functools
import os
import re
import shlex
import shutil
import subprocess
import urllib.request
from enum import Enum
from pathlib import Path, PureWindowsPath
from typing import NamedTuple, Optional, Tuple
from urllib.parse import unquote, urlparse

import packaging.version

try:
    import xdg
    from xdg import BaseDirectory
    from xdg.DesktopEntry import DesktopEntry
except ImportError:
    pass

from ..constants import FileManagerType

_linux_desktop = None


def stock_linux_file_manager() -> str:
    """
    Get stock (system default) file manager for the desktop environment.

    Looks up value only if the desktop environment can be determined.

    All exceptions are raised.

    :return: executable name
    """

    global _linux_desktop
    if _linux_desktop is None:
        _linux_desktop = linux_desktop()

    try:
        desktop = _linux_desktop.name
        desktop = LinuxDesktopFamily.get(desktop) or desktop

        return StandardLinuxFileManager[desktop]
    except KeyError:
        raise Exception(f"The desktop {desktop} is unknown")


def user_linux_file_manager() -> str:
    """
    Determine the file manager for this desktop as set by the user.

    xdg-mime is used to get a .desktop file, from which the executable name is
    extracted. The executable is not examined to see if it is valid or if it even
    exists.

    All exceptions are raised.

    :return: executable name
    """

    xdg_cmd = "xdg-mime query default inode/directory"
    cmd = shlex.split(xdg_cmd)
    try:
        desktop_file: str = subprocess.check_output(cmd, universal_newlines=True)
    except Exception:
        raise Exception(f"Could not determine file manager using {xdg_cmd}")

    # Remove new line character from output
    desktop_file = desktop_file[:-1]
    if desktop_file.endswith(";"):
        desktop_file = desktop_file[:-1]

    for desktop_path in (Path(d) / "applications" for d in BaseDirectory.xdg_data_dirs):
        path = desktop_path / desktop_file
        if path.exists():
            p = str(path)
            try:
                desktop_entry = DesktopEntry(p)
            except Exception:
                raise Exception(f"Could not open desktop entry at {p}")
            try:
                desktop_entry.parse(p)
            except xdg.Exceptions.ParsingError:
                raise Exception(f"Could not parse desktop entry at {p}")
            except Exception:
                raise Exception(f"Desktop entry at {p} might be malformed")

            fm = desktop_entry.getExec()

            # Strip away any extraneous arguments
            fm_cmd = fm.split()[0]
            # Strip away any path information
            fm_cmd = Path(fm_cmd).name
            # Strip away any quotes
            fm_cmd = fm_cmd.replace('"', "")
            fm_cmd = fm_cmd.replace("'", "")

            return fm_cmd

    return ""


def valid_linux_file_manager() -> str:
    """
    Get user's file manager, falling back to using sensible defaults for the particular
    desktop environment.

    All exceptions are caught.

    :return: If the user's default file manager is set and it is known by this module,
     then return it. Otherwise return the stock file manager, if it exists.
    """

    try:
        stock = stock_linux_file_manager()
    except Exception:
        stock = ""

    try:
        user_fm = user_linux_file_manager()
    except Exception:
        user_fm = ""
    else:
        if user_fm not in known_linux_file_managers():
            user_fm = ""

    if not (user_fm or stock):
        return ""

    fm = user_fm if user_fm else stock

    if fm and shutil.which(fm):
        return fm
    else:
        return ""


def known_linux_file_managers() -> Tuple[str, ...]:
    """
    Generate a collection of Linux file managers this module knows about

    :return: tuple of executable names
    """

    return tuple(LinuxFileManagerBehavior.keys())


def linux_file_manager_type(file_manager: str) -> FileManagerType:
    """
    Determine the type of command line arguments the Linux file manager expects
    :param file_manager: executable name
    :return: FileManagerType matching with the executable name, else
     FileManagerType.regular as a fallback
    """

    if file_manager == "caja" and caja_supports_select():
        return FileManagerType.select
    return LinuxFileManagerBehavior.get(file_manager, FileManagerType.regular)


def caja_version() -> Optional[packaging.version.Version]:
    """
    Get the version of Caja via a command line switch
    :return: parsed ver
    """

    try:
        version_string = (
            subprocess.run(["caja", "--version"], stdout=subprocess.PIPE, check=True)
            .stdout.decode()
            .strip()
        )
    except subprocess.CalledProcessError:
        raise Exception("Failed to get version number from caja")

    result = re.search("\d", version_string)
    if result is None:
        return None

    version = version_string[result.start() :]
    return packaging.version.parse(version)


def caja_supports_select() -> bool:
    """
    Determine if caja supports --select command line switch.

    :return: True if caja version is >= version 1.26
    """

    try:
        version = caja_version()
    except Exception:
        return False
    if version is None:
        return False
    return version >= packaging.version.Version("1.26")


def translate_wsl_path(path: str, from_windows_to_wsl: bool) -> str:
    """
    Use the WSL command wslpath to translate between Windows and WSL paths.

    Uses subprocesss. Exceptions not caught.

    :param path: path to convert in string format
    :param from_windows_to_wsl: whether to translate from Windows to WSL (True),
     or WSL to Windows (False)
    :return: the translated path
    """
    arg = "-u" if from_windows_to_wsl else "-w"
    return (
        subprocess.run(["wslpath", arg, path], capture_output=True, check=True)
        .stdout.decode()
        .strip()
    )


def wsl_path_is_for_windows(path_or_uri: str) -> bool:
    """
    When running in WSL, detect if the path being passed is in Windows or Linux

    Reference: https://docs.microsoft.com/en-us/windows/wsl/filesystems

    :param path_or_uri:
    :return:
    """

    if path_or_uri.startswith("file://"):
        # Assume valid URI
        # Look for drive letter, windows style
        if path_or_uri[3:4].isalpha() and path_or_uri[4:5] == ":":  # noqa: SIM114
            return True
        # Look for UNC host name: anything that does not start with a leading /
        elif path_or_uri[7].isalpha():
            return True
        return False
    else:
        drive = PureWindowsPath(path_or_uri).drive
        if drive:
            # C:\
            if drive[0].isalpha() and drive[1] == ":":
                return True
            # UNC share
            if drive[:2] == r"\\":
                return True
        # Assume anything under /mnt is Windows
        return path_or_uri.startswith("/mnt")


class WSLTransformPathURI(NamedTuple):
    is_win_location: bool
    win_uri: str
    win_path: str
    linux_path: str
    is_dir: bool
    exists: bool


def wsl_transform_path_uri(
    path_or_uri: str, generate_win_path: bool
) -> WSLTransformPathURI:
    r"""
    Transforms URI or path into path and URI suitable for working with WSL.

    Detects if working with path or URI, and whether it is POSIX or Windows
    Assumes all paths mounted on /mnt are located in Windows.

    :param path_or_uri: path or URI to examine
    :param generate_win_path: if passed a Linux path, generate path and URI for use in
     Windows. Will do so anyway if path is located in Windows, not the Linux instance.
    :return: Named Tuple containing values in WSLTranformPathURI

    >>> import platform
    >>> assert platform.system() == "Linux"
    >>> r = wsl_transform_path_uri("file:///c:/Program%20Files/Common%20Files/", True)
    >>> r.win_path
    'C:\\Program Files\\Common Files'
    >>> r.is_win_location
    True
    >>> r.linux_path
    '/mnt/c/Program Files/Common Files'
    >>> r.win_uri
    'file:///c:/Program%20Files/Common%20Files/'
    >>> r.is_win_location
    True
    >>> r.is_dir if r.exists else True
    True
    >>> r = wsl_transform_path_uri("file:///c:/Program%20Files/Common%20Files", True)
    >>> r.win_uri
    'file:///c:/Program%20Files/Common%20Files/'
    >>> r = wsl_transform_path_uri("file:///c:/Program Files/Common Files", True)
    >>> r.win_uri
    'file:///c:/Program%20Files/Common%20Files/'
    >>> f ="file:///c:/Program%20Files/Barrier/barrier.conf"
    >>> r = wsl_transform_path_uri(f, True)
    >>> r.win_path
    'C:\\Program Files\\Barrier\\barrier.conf'
    >>> r.is_win_location
    True
    >>> r.linux_path
    '/mnt/c/Program Files/Barrier/barrier.conf'
    >>> r.is_dir if r.exists else False
    False
    >>> r.win_uri
    'file:///c:/Program%20Files/Barrier/barrier.conf'
    >>> from pathlib import Path
    >>> home = Path.home()
    >>> r = wsl_transform_path_uri(f"file://localhost{home}/.bashrc", True)
    >>> r.is_dir if r.exists else False
    False
    >>> r.is_win_location
    False
    >>> r.linux_path # doctest: +ELLIPSIS
    '/home/.../.bashrc'
    >>> r.win_path  # doctest: +ELLIPSIS
    '\\\\...\\...\\home\\...\\.bashrc'
    >>> r.win_uri  # doctest: +ELLIPSIS
    'file://.../.../home/.../.bashrc'
    >>> r = wsl_transform_path_uri(f"file://{home}/.bashrc", True)
    >>> r.is_dir
    False
    >>> r.is_win_location
    False
    >>> r.linux_path # doctest: +ELLIPSIS
    '/home/.../.bashrc'
    >>> r.win_path  # doctest: +ELLIPSIS
    '\\\\...\\...\\home\\...\\.bashrc'
    >>> r.win_uri  # doctest: +ELLIPSIS
    'file://.../.../home/.../.bashrc'
    >>> r = wsl_transform_path_uri(f"file://localhost{home}/my%20file.txt", True)
    >>> r.is_dir if r.exists else False
    False
    >>> f = r'\\wsl.localhost\Ubuntu-20.04\home\damon\my file.txt'
    >>> r.win_path if r.exists else f # doctest: +ELLIPSIS
    '\\\\...\\...\\home\\...\\my file.txt'
    >>> r.is_win_location
    False
    >>> f = 'file://wsl$/Ubuntu-20.04/home/damon/my%20file.txt'
    >>> r.win_uri if r.exists else f # doctest: +ELLIPSIS
    'file://.../.../home/.../my%20file.txt'
    >>> r = wsl_transform_path_uri("file:///etc/fstab", True)
    >>> r.exists
    True
    >>> r.is_dir
    False
    >>> r.linux_path
    '/etc/fstab'
    >>> r.is_win_location
    False
    >>> r.win_path  # doctest: +ELLIPSIS
    '\\\\...\\...\\etc\\fstab'
    >>> r.win_uri  # doctest: +ELLIPSIS
    'file://.../.../etc/fstab'
    >>> r = wsl_transform_path_uri("file:/etc/fstab", True)
    >>> r.exists
    True
    >>> r.is_dir
    False
    >>> r.linux_path
    '/etc/fstab'
    >>> r.is_win_location
    False
    >>> r.win_path  # doctest: +ELLIPSIS
    '\\\\...\\...\\etc\\fstab'
    >>> r.win_uri  # doctest: +ELLIPSIS
    'file://.../.../etc/fstab'
    >>> r = wsl_transform_path_uri("file:///etc", True)
    >>> r.exists
    True
    >>> r.is_dir
    True
    >>> r.linux_path
    '/etc'
    >>> r.is_win_location
    False
    >>> r.win_path  # doctest: +ELLIPSIS
    '\\\\...\\...\\etc'
    >>> r.win_uri  # doctest: +ELLIPSIS
    'file://.../.../etc/'
    >>> r = wsl_transform_path_uri("file:///etc", False)
    >>> r.win_path
    >>> r.win_uri
    >>> r = wsl_transform_path_uri(f"file://{home}/dir with spaces", True)
    >>> r.is_dir if r.exists else True
    True
    >>> r.linux_path  # doctest: +ELLIPSIS
    '/home/.../dir with spaces'
    >>> r.is_win_location
    False
    >>> f = '\\\\wsl.localhost\\openSUSE-Leap-15.3\\home\\damon\\dir with spaces'
    >>> r.win_path if r.exists else f  # doctest: +ELLIPSIS
    '\\\\...\\...\\home\\...\\dir with spaces'
    >>> f = 'file://wsl.localhost/openSUSE-Leap-15.3/home/damon/dir%20with%20spaces/'
    >>> r.win_uri if r.exists else f  # doctest: +ELLIPSIS
    'file://.../.../home/.../dir%20with%20spaces/'
    >>> r = wsl_transform_path_uri("file:///c:/Program%20Files/Common%20Files", True)
    >>> r.is_win_location
    True
    >>> r.win_path
    'C:\\Program Files\\Common Files'
    >>> r.linux_path
    '/mnt/c/Program Files/Common Files'
    >>> r = wsl_transform_path_uri("file:///mnt/c/Program%20Files/", True)
    >>> r.is_win_location
    True
    >>> r.exists
    True
    >>> r.is_dir
    True
    >>> r.win_path
    'C:\\Program Files'
    >>> r.linux_path
    '/mnt/c/Program Files'
    >>> r.win_uri
    'file:///c:/Program%20Files/'
    >>> r = wsl_transform_path_uri(r"C:\Program Files", True)
    >>> r.is_win_location
    True
    >>> r.exists
    True
    >>> r.is_dir
    True
    >>> r.win_path
    'C:\\Program Files'
    >>> r.linux_path
    '/mnt/c/Program Files'
    >>> r.win_uri
    'file:///c:/Program%20Files/'
    >>> import os
    >>> import pwd
    >>> user = pwd.getpwuid(os.getuid())[0]
    >>> f = f"\\\\wsl.localhost\\openSUSE-Leap-15.3\\home\\{user}\\My Photos"
    >>> r = wsl_transform_path_uri(f, True)
    >>> r.is_win_location if r.exists else False
    False
    >>> r.win_path # doctest: +ELLIPSIS
    '\\\\wsl.localhost\\openSUSE-Leap-15.3\\home\\...\\My Photos'
    >>> r.win_uri  # doctest: +ELLIPSIS
    'file://wsl.localhost/openSUSE-Leap-15.3/home/.../My%20Photos...'
    >>> r.linux_path if r.exists else "/home/damon/My Photos"  # doctest: +ELLIPSIS
    '/home/.../My Photos'
    >>> r = wsl_transform_path_uri("/mnt/c/Program Files/", True)
    >>> r.is_win_location
    True
    >>> r.exists
    True
    >>> r.is_dir
    True
    >>> r.win_path
    'C:\\Program Files'
    >>> r.linux_path
    '/mnt/c/Program Files'
    >>> r.win_uri
    'file:///c:/Program%20Files/'
    >>> r = wsl_transform_path_uri(str(home), True)
    >>> r.is_win_location
    False
    >>> r.exists
    True
    >>> r.is_dir
    True
    >>> r.win_path  # doctest: +ELLIPSIS
    '\\\\...\\...\\home\\...'
    >>> r.linux_path  # doctest: +ELLIPSIS
    '/home/...'
    >>> r.win_uri   # doctest: +ELLIPSIS
    'file://.../.../home/.../'
    >>> r = wsl_transform_path_uri(f"{home}/dir with spaces", True)
    >>> r.is_win_location
    False
    >>> r.is_dir if r.exists else True
    True
    >>> r.linux_path  # doctest: +ELLIPSIS
    '/home/.../dir with spaces'
    >>> f = '\\\\wsl.localhost\\openSUSE-Leap-15.3\\home\\damon\\dir with spaces'
    >>> r.win_path if r.exists else f  # doctest: +ELLIPSIS
    '\\\\...\\...\\home\\...\dir with spaces'
    >>> f = 'file://wsl.localhost/openSUSE-Leap-15.3/home/damon/dir%20with%20spaces/'
    >>> r.win_uri  if r.exists else f  # doctest: +ELLIPSIS
    'file://.../.../home/.../dir%20with%20spaces/'
    >>> r = wsl_transform_path_uri(f"{home}/.bashrc", True)
    >>> r.is_win_location
    False
    >>> r.is_dir if r.exists else False
    False
    >>> r.linux_path  # doctest: +ELLIPSIS
    '/home/.../.bashrc'
    >>> r.win_path  # doctest: +ELLIPSIS
    '\\\\...\\...\\home\\...\.bashrc'
    >>> r.win_uri  # doctest: +ELLIPSIS
    'file://.../.../home/.../.bashrc'
    >>> cwd = os.getcwd()
    >>> os.chdir(home)
    >>> r = wsl_transform_path_uri(".bashrc", True)
    >>> r.linux_path   # doctest: +ELLIPSIS
    '/home/.../.bashrc'
    >>> r.win_path  # doctest: +ELLIPSIS
    '\\\\...\\...\\home\\...\.bashrc'
    >>> r.is_dir if r.exists else False
    False
    >>> r.is_win_location
    False
    >>> r.win_uri  # doctest: +ELLIPSIS
    'file://.../.../home/.../.bashrc'
    >>> os.chdir(cwd)
    """

    win_uri = None
    win_path = None
    linux_path = None
    is_dir = None
    exists = None
    if path_or_uri.startswith("file:/"):
        is_win_uri = False
        parsed = urlparse(url=path_or_uri)
        path = unquote(parsed.path)
        netloc = parsed.netloc
        if len(path) > 2 and path[0] == "/" and path[1].isalpha() and path[2] == ":":
            is_win_uri = True
            # Remove first forward slash from e.g. /c:/Program Files
            path = path[1:]
        elif netloc and netloc != "localhost":
            is_win_uri = True

        if is_win_uri:
            win_uri = path_or_uri.replace(" ", "%20")
            win_path = path
            try:
                linux_path = translate_wsl_path(path, from_windows_to_wsl=True)
            except subprocess.CalledProcessError:
                exists = False
        else:
            linux_path = path

    else:
        path = path_or_uri
        if path.startswith("/") and not path.startswith("/mnt"):
            linux_path = path
        else:
            # Path must be either a Windows style path, or a relative path on Posix.
            # First, check if the path is Windows style, e.g. C:\Program Files
            # Note that UNC shares are also considered drives
            drive = PureWindowsPath(path).drive
            is_unc = drive.startswith("\\\\")
            if (drive and drive[0].isalpha() and drive[1] == ":") or is_unc:
                win_path = path
                try:
                    linux_path = translate_wsl_path(path=path, from_windows_to_wsl=True)
                except subprocess.CalledProcessError:
                    exists = False

                # Generate Windows URI
                if is_unc:
                    wuri = urllib.request.pathname2url(path.replace("\\", "/"))
                    win_uri = f"file:{wuri}"
                else:
                    win_uri = wsl_path_to_uri_for_windows_explorer(linux_path)
            else:
                # relative path was passed
                linux_path = str(Path(path).resolve())

    if linux_path is None:
        is_win_location = None
    else:
        lpath = Path(linux_path)
        exists = lpath.exists()

        is_win_location = linux_path.startswith("/mnt/")

        if exists:
            is_dir = lpath.is_dir()
            if generate_win_path or is_win_location:
                try:
                    win_path = translate_wsl_path(
                        path=linux_path, from_windows_to_wsl=False
                    )
                except subprocess.CalledProcessError:
                    exists = False
                if win_path and not win_uri:
                    if not is_win_location:
                        wuri = urllib.request.pathname2url(win_path.replace("\\", "/"))
                        win_uri = f"file:{wuri}"
                    else:
                        win_uri = wsl_path_to_uri_for_windows_explorer(linux_path)

    if is_dir:
        if linux_path is not None and linux_path[-1] == "/":
            linux_path = linux_path[:-1]
        if win_path is not None and win_path[-1] == "\\":
            win_path = win_path[:-1]
        if win_uri is not None and win_uri[-1] != "/":
            win_uri = f"{win_uri}/"

    return WSLTransformPathURI(
        is_win_location=is_win_location,
        win_uri=win_uri,
        win_path=win_path,
        linux_path=linux_path,
        is_dir=is_dir,
        exists=exists,
    )


def wsl_path_to_uri_for_windows_explorer(path: str) -> str:
    r"""
    Convert a path to a URI accepted by Windows Explorer.

    Windows URIs are different from Linux URIs. As Wikipedia points out, Window
    specifies file:///c:/path/to/the%20file.txt
    (note the three slashes after file:), and
    file://hostname/path/to/the%20file.txt

    :param path: path format like '/mnt/c/some/path', '/home/user', 'C:\some\path'
    :return: a file URI accepted by Windows Explorer
    """

    assert not path.startswith("\\\\")
    assert path.startswith("/mnt/")

    path = urllib.request.pathname2url(path)
    # Remove the /mnt portion, keep the drive letter, and insert a colon
    return f"file://{path[4:6]}:{path[6:]}"


class LinuxDesktop(Enum):
    gnome = 1
    unity = 2
    cinnamon = 3
    kde = 4
    xfce = 5
    mate = 6
    lxde = 7
    lxqt = 8
    ubuntugnome = 9
    popgnome = 10
    deepin = 11
    zorin = 12
    ukui = 13  # Kylin
    pantheon = 14
    enlightenment = 15
    wsl = 16
    wsl2 = 17
    cutefish = 18
    lumina = 19
    unknown = 20


LinuxDesktopHumanize = dict(
    gnome="Gnome",
    unity="Unity",
    cinnamon="Cinnamon",
    kde="KDE",
    xfce="XFCE",
    mate="Mate",
    lxde="LXDE",
    lxqt="LxQt",
    ubuntugnome="Ubuntu Gnome",
    popgnome="Pop Gnome",
    deepin="Deepin",
    zorin="Zorin",
    ukui="UKUI",
    pantheon="Pantheon",
    enlightenment="Enlightenment",
    wsl="WSL1",
    wsl2="WSL2",
    cutefish="Cutefish",
    lumina="Lumina",
    unknown="Unknown",
)


LinuxDesktopFamily = dict(
    ubuntugnome="gnome",
    popgnome="gnome",
    zorin="gnome",
    unity="gnome",
)


StandardLinuxFileManager = dict(
    gnome="nautilus",
    kde="dolphin",
    cinnamon="nemo",
    mate="caja",
    xfce="thunar",
    lxde="pcmanfm",
    lxqt="pcmanfm-qt",
    deepin="dde-file-manager",
    pantheon="io.elementary.files",
    ukui="peony",
    enlightenment="pcmanfm",
    wsl="explorer.exe",
    wsl2="explorer.exe",
    cutefish="cutefish-filemanager",
    lumina="lumina-fm",
)


LinuxFileManagerBehavior = dict(
    nautilus=FileManagerType.select,
    dolphin=FileManagerType.select,
    caja=FileManagerType.dir_only_uri,
    thunar=FileManagerType.dir_only_uri,
    nemo=FileManagerType.regular,
    pcmanfm=FileManagerType.dir_only_uri,
    peony=FileManagerType.show_items,
    index=FileManagerType.dir_only_uri,
    doublecmd=FileManagerType.dual_panel,
    krusader=FileManagerType.dir_only_uri,
    spacefm=FileManagerType.dir_only_uri,
    fman=FileManagerType.dual_panel,
)
LinuxFileManagerBehavior["pcmanfm-qt"] = FileManagerType.dir_only_uri
LinuxFileManagerBehavior["dde-file-manager"] = FileManagerType.show_item
LinuxFileManagerBehavior["io.elementary.files"] = FileManagerType.regular
LinuxFileManagerBehavior["cutefish-filemanager"] = FileManagerType.dir_only_uri
LinuxFileManagerBehavior["lumina-fm"] = FileManagerType.dir_only_uri


def wsl_version() -> Optional[LinuxDesktop]:
    with open("/proc/version") as f:
        p = f.read()
    if p.find("microsoft") > 0 and p.find("WSL2"):
        return LinuxDesktop.wsl2
    if p.find("Microsoft") > 0:
        return LinuxDesktop.wsl
    return None


def detect_wsl() -> bool:
    with open("/proc/version") as f:
        p = f.read()
    return p.lower().find("microsoft") > 0


@functools.lru_cache(maxsize=None)
def linux_desktop() -> LinuxDesktop:
    """
    Determine Linux desktop environment

    :return: enum representing desktop environment, Desktop.unknown if unknown.
    """

    try:
        env = os.getenv("XDG_CURRENT_DESKTOP").lower()
    except AttributeError:
        wsl = wsl_version()
        if wsl is not None:
            return wsl
        else:
            raise Exception("The value for XDG_CURRENT_DESKTOP is not set")

    if env in ("unity:unity7", "unity:unity7:ubuntu"):
        env = "unity"
    elif env == "x-cinnamon":
        env = "cinnamon"
    elif env == "ubuntu:gnome":
        env = "ubuntugnome"
    elif env == "pop:gnome":
        env = "popgnome"
    elif env in ("gnome-classic:gnome", "budgie:gnome"):
        env = "gnome"
    elif env == "zorin:gnome":
        env = "zorin"

    try:
        return LinuxDesktop[env]
    except KeyError:
        raise Exception(f"The desktop environment {env} is unknown")


def linux_desktop_humanize(desktop: LinuxDesktop) -> str:
    """
    Make LinuxDesktop name human readable.
    :return: desktop name spelled out
    """

    return LinuxDesktopHumanize[desktop.name]



# Copyright (c) 2016-2024 Damon Lynch
# SPDX - License - Identifier: MIT

"""
Parse command line arguments
"""

import importlib.metadata
from argparse import ArgumentParser, HelpFormatter


def package_metadata():
    """
    Get Python package metadata

    :return: version number and package summary
    """

    try:
        version = importlib.metadata.version("show-in-file-manager")
    except Exception:
        version = "Unknown version"
        summary = (
            "Platform independent Python module to open the system file manager "
            "and optionally select files in it "
        )

    else:
        metadata = importlib.metadata.metadata("show-in-file-manager")
        summary = metadata["summary"]

    return version, summary


def get_parser(formatter_class=HelpFormatter) -> ArgumentParser:
    """
    Parse command line options for this script

    :param formatter_class: one of 4 argparse formatting classes
    :return: argparse.ArgumentParser
    """

    version, summary = package_metadata()

    parser = ArgumentParser(
        prog="showinfilemanager", description=summary, formatter_class=formatter_class
    )

    parser.add_argument("--version", action="version", version=f"%(prog)s {version}")

    parser.add_argument("-f", "--file-manager", help="file manager to run")

    parser.add_argument(
        "-s",
        "--select-folder",
        action="store_true",
        help="select folder instead of displaying its contents",
    )

    parser.add_argument(
        "--verbose", action="store_true", help="display command being run to stdout"
    )

    parser.add_argument(
        "--debug", action="store_true", help="output debugging information to stdout"
    )

    parser.add_argument(
        "path", nargs="*", help="zero or more URIs or paths of files or directories"
    )

    return parser



# Copyright (c) 2021 Damon Lynch
# SPDX - License - Identifier: MIT

import contextlib
from pathlib import Path
from typing import List, Optional

with contextlib.suppress(ImportError):
    from win32com.shell import shell

from ..constants import FileManagerType
from .tools import (
    directories_and_their_files,
    file_uri_to_path,
    is_uri,
    path_to_file_uri,
)

WindowsFileManagerBehavior = {}
WindowsFileManagerBehavior["doublecmd.exe"] = FileManagerType.dual_panel
WindowsFileManagerBehavior["fman.exe"] = FileManagerType.dual_panel


def windows_file_manager_type(file_manager: str) -> FileManagerType:
    """
    Determine the type of command line arguments the Windows file manager expects
    :param file_manager: executable name
    :return: FileManagerType matching with the executable name, else
     FileManagerType.regular as a fallback
    """

    if file_manager == "explorer.exe":
        return FileManagerType.win_select
    return WindowsFileManagerBehavior.get(file_manager, FileManagerType.regular)


def parse_command_line_arguments(path_or_uri: List[str]) -> List[str]:
    """
    Convert any glob component in the filename component of Windows paths, which the
    Windows shell does not do itself

    :param path_or_uri: list of paths or URIs
    :return: list of paths or URIs with resolved paths and no glob components
    """

    paths = []
    for pu in path_or_uri:
        if is_uri(pu):
            uri = pu
            path = Path(file_uri_to_path(uri=uri))
        else:
            uri = None
            path = Path(pu)
        if not path.is_dir():
            for globbed_pu in path.parent.resolve().glob(path.name):
                if uri:
                    paths.append(path_to_file_uri(str(globbed_pu)))
                else:
                    paths.append(str(globbed_pu))
        else:
            if uri:
                paths.append(path_to_file_uri(str(path)))
            else:
                paths.append(str(path.resolve()))

    return paths


def launch_file_explorer(paths: List[str], verbose: Optional[bool] = False) -> None:
    """
    Open Windows File Explorer, selecting files in their folders

    Inspired by https://mail.python.org/pipermail/python-win32/2012-September/012531.html
    and http://mail.python.org/pipermail/python-win32/2012-September/012533.html
    """

    folder_contents = directories_and_their_files(paths)

    if folder_contents:
        desktop = shell.SHGetDesktopFolder()
        for folder in folder_contents:
            folder_pidl = shell.SHILCreateFromPath(folder, 0)[0]
            shell_folder = desktop.BindToObject(
                folder_pidl, None, shell.IID_IShellFolder
            )
            win_items = {
                desktop.GetDisplayNameOf(item, 0): item for item in shell_folder
            }
            to_select = [
                win_items[file] for file in folder_contents[folder] if file in win_items
            ]
            if verbose:
                files = '", "'.join(folder_contents[folder])
                if files:
                    files = f'"{files}"'
                print(
                    "Executing Windows shell to open file "
                    f'explorer at "{folder}", selecting {files}'
                )
            shell.SHOpenFolderAndSelectItems(folder_pidl, to_select, 0)



# Copyright (c) 2021 Damon Lynch
# SPDX - License - Identifier: MIT

# ruff: noqa: F401

from showinfm.constants import cannot_open_uris, single_file_only
from showinfm.showinfm import (
    show_in_file_manager,
    stock_file_manager,
    user_file_manager,
    valid_file_manager,
)
from showinfm.system.linux import LinuxDesktop, linux_desktop, linux_desktop_humanize



# Copyright (c) 2016-2024 Damon Lynch
# SPDX - License - Identifier: MIT

"""
Show in File Manager

Open the system file manager and optionally select files in it.
"""

import os
import platform
import shlex
import shutil
import subprocess
import sys
import urllib.parse
from pathlib import Path
from typing import List, Optional, Sequence, Union

from .argumentsparse import get_parser
from .constants import FileManagerType, Platform, single_file_only
from .system import current_platform, is_wsl, is_wsl1, is_wsl2, linux, tools, windows

_valid_file_manager_probed = False
_valid_file_manager = None
_valid_file_manager_type = None


def stock_file_manager() -> str:
    """
    Get stock file manager for this operating system / desktop.

    On Windows the default is `explorer.exe`. On Linux the first step
    is to determine which desktop is running, and from that lookup its
    default file manager. On macOS, the default is finder, accessed
    via the command 'open'.

    Exceptions are not caught.

    :return: executable name
    """

    if current_platform == Platform.windows or is_wsl1:
        file_manager = "explorer.exe"
    elif current_platform == Platform.linux:
        file_manager = linux.stock_linux_file_manager()
    elif current_platform == Platform.macos:
        file_manager = "open"
    else:
        raise NotImplementedError

    assert shutil.which(file_manager) is not None
    return file_manager


def user_file_manager() -> str:
    """
    Get the file manager as set by the user.

    The file manager executable is tested to see if it exists.

    Exceptions are not caught.

    :return: executable name
    """

    if current_platform == Platform.windows or is_wsl1:
        file_manager = "explorer.exe"
    elif current_platform == Platform.linux:
        file_manager = linux.user_linux_file_manager()
    elif current_platform == Platform.macos:
        file_manager = "open"
    else:
        raise NotImplementedError

    assert shutil.which(file_manager) is not None
    return file_manager


def valid_file_manager() -> str:
    """
    Get user's file manager, falling back to using sensible defaults.

    The user's choice of file manager is the default choice. However, this is
    not always set correctly. On Linux, it most likely is because the user's
    distro has not correctly set the default file manager. If the user's choice
    is unrecognized by this package, then reject it and choose the standard file
    manager for the detected desktop environment.

    All exceptions are caught, except those if this platform is not supported by
    this package.

    :return: If the user's default file manager is set and it is recognized
     as valid by this package, then return it. Otherwise return the stock file
     manager, if it exists.
    """

    if current_platform == Platform.windows or is_wsl1:
        file_manager = "explorer.exe"
    elif current_platform == Platform.linux:
        file_manager = linux.valid_linux_file_manager()
    elif current_platform == Platform.macos:
        file_manager = "open"
    else:
        raise NotImplementedError
    return file_manager


def show_in_file_manager(
    path_or_uri: Optional[Union[str, Sequence[str]]] = None,
    open_not_select_directory: Optional[bool] = True,
    file_manager: Optional[str] = None,
    allow_conversion: bool = True,
    verbose: bool = False,
    debug: bool = False,
) -> None:
    """
    Open the file manager and show zero or more directories or files in it.

    The path_or_uri is a sequence of items, or a single item. An item can
    be regular path, or a URI.

    On non-Windows platforms, regular paths will be converted to URIs when
    passed as command line arguments to the file manager, because some file
    managers do not handle regular paths correctly. However, URIs will be
    convereted to paths to handle file managers that do not accepts URIs.

    On Windows, Explorer is called using the Win32 API.

    On WSL1, all paths are opened using Windows Explorer. URIs and can be
    specified using Linux or Windows formats. All formats are automatically
    converted to use the Windows URI format.

    WSL2 functions the same as WSL1, except if the WSL2 instance has a Linux
    file manager installed. On these systems, if a path on Linux is
    specified, that file manager will be used instead of Windows Explorer. You
    can override this default behavior by using the parameter file_manager.

    The most common use of this function is to call it without specifying
    the file manager to use, which defaults to the value returned by
    valid_file_manager()

    For file managers unable to select files to display, the file manager
    will instead display the contents of the path.

    For file managers that can handle file selections, but only one at time,
    multiple file manager windows will be opened.

    If you specify a file manager executable and this package does not
    recognize it, it will be called with the files as the only command line
    arguments.

    :param path_or_uri: zero or more files or directories to open, specified
     as a single URI or valid path, or a sequence of URIs/paths.
    :param open_not_select_directory: if the URI or path is a directory and
     not a file, open the directory itself in the file manager, rather than
     selecting it and displaying it in its parent directory.
    :param file_manager: executable name to use. If not specified, then
     valid_file_manager() will determine which file manager to use.
    :param allow_conversion: allow this function to automatically convert paths
     and URIs to the format needed by the file manager that will be called. Set
     to False if passing non-standard URIs. Ignored when running under WSL.
    :param verbose: if True print command to be executed before launching
     it
    :param debug: if True print debugging information to stderr
    """

    global _valid_file_manager
    global _valid_file_manager_type

    file_manager_specified = file_manager is not None

    if not file_manager:
        _set_valid_file_manager()
        file_manager_type = _valid_file_manager_type
        file_manager = _valid_file_manager
    else:
        try:
            file_manager_type = _file_manager_type(file_manager)
        except Exception:
            file_manager_type = None

    if not (file_manager or is_wsl2):
        # We are not running WSL2 and there is no file manager: there is nothing
        # to be done
        return

    # Used for directories when open_not_select_directory is True
    directories = []
    # Used for paths that will be opened in Windows explorer called from WSL2
    wsl_windows_paths = []
    wsl_windows_directories = []

    if not path_or_uri:
        if current_platform == Platform.macos:
            # macOS finder requires a path to be able to launch it from the command line
            path_or_uri = "file:///"
        else:
            arg = ""
            uris_and_paths = []
    if path_or_uri:
        if isinstance(path_or_uri, str):
            # turn the single path / URI into a Tuple
            path_or_uri = (path_or_uri,)

        uris_and_paths = []
        for pu in path_or_uri:
            # Were we passed a URI or simple path?
            if pu:
                uri = path = ""
                # target_platform = current_platform
                if is_wsl:
                    # Running WSL1 or WSL2. Is the path on the Windows file system, or
                    # alternately is Windows explorer going to be used to view the
                    # files? Also, what kind of path or URI has been passed?
                    require_win_path = file_manager == "explorer.exe"
                    wsl_details = linux.wsl_transform_path_uri(pu, require_win_path)
                    if not wsl_details.exists:
                        continue
                    use_windows_explorer_via_wsl = (
                        wsl_details.is_win_location and not file_manager_specified
                    ) or file_manager == "explorer.exe"
                    if use_windows_explorer_via_wsl:
                        if debug:
                            print(
                                f"Converted '{pu}' to '{wsl_details.win_uri}'",
                                file=sys.stderr,
                            )
                        if not (wsl_details.is_dir and open_not_select_directory):
                            wsl_windows_paths.append(wsl_details.win_uri)
                        else:
                            wsl_windows_directories.append(wsl_details.win_uri)
                        continue
                    else:
                        if tools.filemanager_requires_path(file_manager=file_manager):
                            path = Path(wsl_details.linux_path).resolve()
                            uri = None
                        else:
                            path = None
                            uri = Path(wsl_details.linux_path).resolve().as_uri()
                else:
                    if tools.is_uri(pu):
                        if (
                            tools.filemanager_requires_path(file_manager=file_manager)
                            and allow_conversion
                        ):
                            # Convert URI to regular path
                            uri = None
                            path = Path(path or tools.file_uri_to_path(pu))
                        else:
                            uri = pu
                            path = None
                    else:
                        if (
                            tools.filemanager_requires_path(file_manager=file_manager)
                            or not allow_conversion
                        ):
                            path = Path(pu)
                            uri = None
                        else:
                            uri = Path(pu).resolve().as_uri()
                            path = None

                if file_manager_type == FileManagerType.dir_only_uri:
                    assert current_platform != Platform.windows
                    # Show only the directory: do not attempt to select the file,
                    # because the file manager cannot handle it.
                    if uri:
                        # Do not use tools.file_url_to_path() here, because we need the
                        # parse_result, and file_url_to_path() assumes file:// URIs.
                        # In any case, this code block is not run under Windows, so
                        # there is no need to use tools.file_url_to_path() to handle the
                        # file:/// case that urllib.parse.urlparse fails with.
                        parse_result = urllib.parse.urlparse(uri)
                        path = Path(parse_result.path)

                    if not (path.is_dir() and open_not_select_directory):
                        path = path.parent
                    if uri:
                        uri = urllib.parse.urlunparse(
                            parse_result._replace(path=str(path))
                        )
                    else:
                        path = tools.quote_path(path=path)
                    uris_and_paths.append(uri or str(path))
                else:
                    # whether to open the directory, or
                    # select it (depends on file manager capabilities and option
                    # open_not_select_directory):
                    open_directory = False

                    if (
                        open_not_select_directory
                        and file_manager_type != FileManagerType.dual_panel
                        or file_manager_type == FileManagerType.regular
                    ):
                        if uri:
                            path = tools.file_uri_to_path(uri=uri)
                            path = Path(path)
                            open_directory = path.is_dir()
                        else:
                            open_directory = path.is_dir()
                        if open_directory:
                            if (
                                file_manager_type == FileManagerType.regular
                                and not open_not_select_directory
                            ):
                                # This type of file manger cannot select directories,
                                # because it provides no mechanism
                                # to distinguish between selecting and opening a
                                # directory.
                                # So open the parent instead.
                                path = path.parent
                                if uri:
                                    uri = path.as_uri()
                            if uri is None:
                                path = tools.quote_path(path=path)
                            directories.append(uri or str(path))
                    if not open_directory:
                        if uri is None and file_manager != "explorer.exe":
                            path = tools.quote_path(path=path)
                        uris_and_paths.append(uri or str(path))

        arg = ""
        if file_manager_type == FileManagerType.win_select:
            arg = "/select,"  # no trailing space is necessary on Windows
        elif file_manager_type == FileManagerType.select:
            arg = "--select "  # trailing space is necessary
        elif file_manager_type == FileManagerType.show_item:
            arg = "--show-item "  # trailing space is necessary
        elif file_manager_type == FileManagerType.show_items:
            arg = "--show-items "  # trailing space is necessary
        elif file_manager_type == FileManagerType.reveal:
            arg = "--reveal "  # trailing space is necessary

    if (
        current_platform == Platform.windows
        and not is_wsl
        and file_manager == "explorer.exe"
    ):
        if uris_and_paths:
            windows.launch_file_explorer(uris_and_paths, verbose)
        for d in directories:
            if verbose:
                print("Executing Windows shell to open", d)
            os.startfile(d)
    else:
        if uris_and_paths:
            # Some file managers must be passed only one or zero paths / URIs
            if file_manager not in single_file_only:
                uris_and_paths = [" ".join(uris_and_paths)]

            _launch_file_manager(
                uris_or_paths=uris_and_paths,
                arg=arg,
                file_manager=file_manager,
                verbose=verbose,
            )
        if directories:
            if file_manager not in single_file_only:
                directories = [" ".join(directories)]
            _launch_file_manager(
                uris_or_paths=directories,
                arg="",
                file_manager=file_manager,
                verbose=verbose,
            )
        if wsl_windows_paths:
            _launch_file_manager(
                uris_or_paths=wsl_windows_paths,
                arg="/select,",
                file_manager="explorer.exe",
                verbose=verbose,
            )
        if wsl_windows_directories:
            _launch_file_manager(
                uris_or_paths=wsl_windows_directories,
                arg="",
                file_manager="explorer.exe",
                verbose=verbose,
            )

    if (
        not uris_and_paths
        and not directories
        and not wsl_windows_paths
        and not wsl_windows_directories
    ):
        _launch_file_manager(
            uris_or_paths=[""],
            arg="",
            file_manager=file_manager,
            verbose=verbose,
        )


def _launch_file_manager(
    uris_or_paths: List[str], arg: str, file_manager: str, verbose: bool
) -> None:
    """
    Launch the file manager

    :param uris_or_paths: list of URIs, or a list of a single empty string
    :param arg: arg to pass the file manager
    :param file_manager: file manager executable name
    :param verbose: if True print command to be executed before launching
     it
    """

    for u in uris_or_paths:
        cmd = f"{file_manager} {arg}{u}"
        if verbose:
            print("Executing", cmd)
        # Do not check current_platform here, it makes no sense
        args = shlex.split(cmd) if platform.system() != "Windows" else cmd
        proc = subprocess.Popen(args)
        if is_wsl2 and file_manager == "explorer.exe":
            proc.wait()


def _file_manager_type(fm: str) -> FileManagerType:
    """
    Determine file manager type via the executable name
    :param fm: executable name
    :return:
    """

    if current_platform == Platform.windows or is_wsl1:
        return windows.windows_file_manager_type(fm)
    elif current_platform == Platform.linux:
        return linux.linux_file_manager_type(fm)
    elif current_platform == Platform.macos:
        return FileManagerType.reveal
    else:
        raise NotImplementedError


def _set_valid_file_manager() -> None:
    """
    Set module level global variables to set a valid file manager for this user in this
    desktop environment.
    """

    global _valid_file_manager_probed
    global _valid_file_manager
    global _valid_file_manager_type

    if not _valid_file_manager_probed:
        fm = valid_file_manager()
        if fm:
            _valid_file_manager = fm
            _valid_file_manager_type = _file_manager_type(fm)
        _valid_file_manager_probed = True


class Diagnostics:
    """
    Collect basic diagnostics information for this package.
    """

    def __init__(self) -> None:
        try:
            self.stock_file_manager = stock_file_manager()
        except Exception as e:
            self.stock_file_manager = str(e)
        try:
            self.user_file_manager = user_file_manager()
        except Exception as e:
            self.user_file_manager = str(e)
        try:
            self.valid_file_manager = valid_file_manager()
        except Exception as e:
            self.valid_file_manager = str(e)

        if current_platform == Platform.linux:
            try:
                self.desktop = linux.linux_desktop()
            except Exception:
                self.desktop = linux.LinuxDesktop.unknown
        else:
            self.desktop = ""

        if is_wsl:
            if is_wsl2:
                self.wsl_version = "2"
            else:
                self.wsl_version = "1"
        else:
            self.wsl_version = ""

    def __str__(self) -> str:
        desktop = f"Linux Desktop: {self.desktop.name}\n" if self.desktop else ""
        wsl = f"WSL version {self.wsl_version}\n" if self.wsl_version else ""
        file_managers = "Stock: {}\nUser's choice: {}\nValid: {}".format(
            self.stock_file_manager, self.user_file_manager, self.valid_file_manager
        )
        return desktop + wsl + file_managers


def main() -> None:
    parser = get_parser()

    args = parser.parse_args()

    verbose = args.verbose
    debug = args.debug
    if debug:
        print(Diagnostics())
        verbose = True

    if current_platform == Platform.windows and not is_wsl:
        path_or_uri = windows.parse_command_line_arguments(args.path)
    else:
        path_or_uri = args.path

    open_not_select_directory = not args.select_folder

    try:
        show_in_file_manager(
            file_manager=args.file_manager,
            path_or_uri=path_or_uri,
            verbose=verbose,
            open_not_select_directory=open_not_select_directory,
            debug=debug,
        )
    except Exception as e:
        sys.stderr.write(str(e))
        if args.debug:
            raise



# Source: https://gist.github.com/mnot/138549
# Converted from Python2 using https://www.pythonconverter.com/
# Not formatted using black

"""
Regex for URIs
These regex are directly derived from the collected ABNF in RFC3986
(except for DIGIT, ALPHA and HEXDIG, defined by RFC2234).
Additional regex are defined to validate the following schemes according to
their respective specifications:
  - http
  - https
  - file
  - data
  - gopher
  - ws
  - wss
  - mailto

See FIXME for areas that still need work.
They should be processed with re.VERBOSE.
"""

__license__ = """
Copyright (c) 2009-2015 Mark Nottingham (code portions)
Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:
The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.
THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
THE SOFTWARE.
"""

### basics -

DIGIT = r"[\x30-\x39]"

ALPHA = r"[\x41-\x5A\x61-\x7A]"

HEXDIG = r"[\x30-\x39A-Fa-f]"

DQUOTE = r"\""

#   pct-encoded   = "%" HEXDIG HEXDIG
pct_encoded = r" %% %(HEXDIG)s %(HEXDIG)s" % locals()

#   unreserved    = ALPHA / DIGIT / "-" / "." / "_" / "~"
unreserved = r"(?: %(ALPHA)s | %(DIGIT)s | \- | \. | _ | ~ )" % locals()

#   gen-delims    = ":" / "/" / "?" / "#" / "[" / "]" / "@"
gen_delims = r"(?: : | / | \? | \# | \[ | \] | @ )"

#   sub-delims    = "!" / "$" / "&" / "'" / "(" / ")"
#                 / "*" / "+" / "," / ";" / "="
sub_delims = r"""(?: ! | \$ | & | ' | \( | \) |
                     \* | \+ | , | ; | = )"""

#   pchar         = unreserved / pct-encoded / sub-delims / ":" / "@"
pchar = r"(?: %(unreserved)s | %(pct_encoded)s | %(sub_delims)s | : | @ )" % locals()

#   reserved      = gen-delims / sub-delims
reserved = r"(?: %(gen_delims)s | %(sub_delims)s )" % locals()

### scheme

#   scheme        = ALPHA *( ALPHA / DIGIT / "+" / "-" / "." )
scheme = r"%(ALPHA)s (?: %(ALPHA)s | %(DIGIT)s | \+ | \- | \. )*" % locals()

### authority

#   dec-octet     = DIGIT                 ; 0-9
#                 / %x31-39 DIGIT         ; 10-99
#                 / "1" 2DIGIT            ; 100-199
#                 / "2" %x30-34 DIGIT     ; 200-249
#                 / "25" %x30-35          ; 250-255
dec_octet = r"""(?: %(DIGIT)s |
                    [\x31-\x39] %(DIGIT)s |
                    1 %(DIGIT)s{2} |
                    2 [\x30-\x34] %(DIGIT)s |
                    25 [\x30-\x35]
                )
""" % locals()

#  IPv4address   = dec-octet "." dec-octet "." dec-octet "." dec-octet
IPv4address = r"%(dec_octet)s \. %(dec_octet)s \. %(dec_octet)s \. %(dec_octet)s" % locals()

#  h16           = 1*4HEXDIG
h16 = r"(?: %(HEXDIG)s ){1,4}" % locals()

#  ls32          = ( h16 ":" h16 ) / IPv4address
ls32 = r"(?: (?: %(h16)s : %(h16)s ) | %(IPv4address)s )" % locals()

#   IPv6address   =                            6( h16 ":" ) ls32
#                 /                       "::" 5( h16 ":" ) ls32
#                 / [               h16 ] "::" 4( h16 ":" ) ls32
#                 / [ *1( h16 ":" ) h16 ] "::" 3( h16 ":" ) ls32
#                 / [ *2( h16 ":" ) h16 ] "::" 2( h16 ":" ) ls32
#                 / [ *3( h16 ":" ) h16 ] "::"    h16 ":"   ls32
#                 / [ *4( h16 ":" ) h16 ] "::"              ls32
#                 / [ *5( h16 ":" ) h16 ] "::"              h16
#                 / [ *6( h16 ":" ) h16 ] "::"
IPv6address = r"""(?:                         (?: %(h16)s : ){6} %(ls32)s |
                                           :: (?: %(h16)s : ){5} %(ls32)s |
       (?:                      %(h16)s )? :: (?: %(h16)s : ){4} %(ls32)s |
       (?: (?: %(h16)s : ){0,1} %(h16)s )? :: (?: %(h16)s : ){3} %(ls32)s |
       (?: (?: %(h16)s : ){0,2} %(h16)s )? :: (?: %(h16)s : ){2} %(ls32)s |
       (?: (?: %(h16)s : ){0,3} %(h16)s )? ::     %(h16)s :      %(ls32)s |
       (?: (?: %(h16)s : ){0,4} %(h16)s )? ::                    %(ls32)s |
       (?: (?: %(h16)s : ){0,5} %(h16)s )? ::                    %(h16)s  |
       (?: (?: %(h16)s : ){0,6} %(h16)s )? ::
)
""" % locals()

#   IPvFuture     = "v" 1*HEXDIG "." 1*( unreserved / sub-delims / ":" )
IPvFuture = r"v %(HEXDIG)s+ \. (?: %(unreserved)s | %(sub_delims)s | : )+" % locals()

#   IP-literal    = "[" ( IPv6address / IPvFuture  ) "]"
IP_literal = r"\[ (?: %(IPv6address)s | %(IPvFuture)s ) \]" % locals()

#   reg-name      = *( unreserved / pct-encoded / sub-delims )
reg_name = r"(?: %(unreserved)s | %(pct_encoded)s | %(sub_delims)s )*" % locals()

#   userinfo      = *( unreserved / pct-encoded / sub-delims / ":" )
userinfo = r"(?: %(unreserved)s | %(pct_encoded)s | %(sub_delims)s | : )*" % locals()

#   host          = IP-literal / IPv4address / reg-name
host = r"(?: %(IP_literal)s | %(IPv4address)s | %(reg_name)s )" % locals()

#   port          = *DIGIT
port = r"(?: %(DIGIT)s )*" % locals()

#   authority     = [ userinfo "@" ] host [ ":" port ]
authority = r"(?: %(userinfo)s @)? %(host)s (?: : %(port)s)?" % locals()

### Path

#   segment       = *pchar
segment = r"%(pchar)s*" % locals()

#   segment-nz    = 1*pchar
segment_nz = r"%(pchar)s+" % locals()

#   segment-nz-nc = 1*( unreserved / pct-encoded / sub-delims / "@" )
#                 ; non-zero-length segment without any colon ":"
segment_nz_nc = r"(?: %(unreserved)s | %(pct_encoded)s | %(sub_delims)s | @ )+" % locals()

#   path-abempty  = *( "/" segment )
path_abempty = r"(?: / %(segment)s )*" % locals()

#   path-absolute = "/" [ segment-nz *( "/" segment ) ]
path_absolute = r"/ (?: %(segment_nz)s (?: / %(segment)s )* )?" % locals()

#   path-noscheme = segment-nz-nc *( "/" segment )
path_noscheme = r"%(segment_nz_nc)s (?: / %(segment)s )*" % locals()

#   path-rootless = segment-nz *( "/" segment )
path_rootless = r"%(segment_nz)s (?: / %(segment)s )*" % locals()

#   path-empty    = 0<pchar>
path_empty = r""  ### FIXME

#   path          = path-abempty    ; begins with "/" or is empty
#                 / path-absolute   ; begins with "/" but not "//"
#                 / path-noscheme   ; begins with a non-colon segment
#                 / path-rootless   ; begins with a segment
#                 / path-empty      ; zero characters
path = r"""(?: %(path_abempty)s |
               %(path_absolute)s |
               %(path_noscheme)s |
               %(path_rootless)s |
               %(path_empty)s
            )
""" % locals()

### Query and Fragment

#   query         = *( pchar / "/" / "?" )
query = r"(?: %(pchar)s | / | \? )*" % locals()

#   fragment      = *( pchar / "/" / "?" )
fragment = r"(?: %(pchar)s | / | \? )*" % locals()

### URIs

#   hier-part     = "//" authority path-abempty
#                 / path-absolute
#                 / path-rootless
#                 / path-empty
hier_part = r"""(?: (?: // %(authority)s %(path_abempty)s ) |
                    %(path_absolute)s |
                    %(path_rootless)s |
                    %(path_empty)s
                )
""" % locals()

#   relative-part = "//" authority path-abempty
#                 / path-absolute
#                 / path-noscheme
#                 / path-empty
relative_part = r"""(?: (?: // %(authority)s %(path_abempty)s ) |
                        %(path_absolute)s |
                        %(path_noscheme)s |
                        %(path_empty)s
                    )
""" % locals()

#   relative-ref  = relative-part [ "?" query ] [ "#" fragment ]
relative_ref = r"%(relative_part)s (?: \? %(query)s)? (?: \# %(fragment)s)?" % locals()

#   URI           = scheme ":" hier-part [ "?" query ] [ "#" fragment ]
URI = r"(?: %(scheme)s : %(hier_part)s (?: \? %(query)s )? (?: \# %(fragment)s )? )" % locals()

#   URI-reference = URI / relative-ref
URI_reference = r"(?: %(URI)s | %(relative_ref)s )" % locals()

#   absolute-URI  = scheme ":" hier-part [ "?" query ]
absolute_URI = r"(?: %(scheme)s : %(hier_part)s (?: \? %(query)s )? )" % locals()

### HTTP[S] - RFC7230

# http-URI = "http:" "//" authority path-abempty [ "?" query ]
#             [ "#" fragment ]

http_URI = r"(?: http: // %(authority)s %(path_abempty)s (?: \? %(query)s )? (?: \# %(fragment)s )? )" % locals()

# https-URI = "https:" "//" authority path-abempty [ "?" query ]
#              [ "#" fragment ]

https_URI = r"(?: https: // %(authority)s %(path_abempty)s (?: \? %(query)s )? (?: \# %(fragment)s )? )" % locals()

### WS[S] - RFC6455

# ws-URI = "ws:" "//" host [ ":" port ] path [ "?" query ]

ws_URI = r"(?: ws: // %(host)s (?: : %(port)s )? %(path)s (?: \? %(query)s )? )" % locals()

# wss-URI = "wss:" "//" host [ ":" port ] path [ "?" query ]

wss_URI = r"(?: wss: // %(host)s (?: : %(port)s )? %(path)s (?: \? %(query)s )? )" % locals()

### mailto - RFC6068

# some-delims  = "!" / "$" / "'" / "(" / ")" / "*"
#            / "+" / "," / ";" / ":" / "@"

some_delims = r"""(?: ! | \$ | ' | \( | \) | \*
                      \+ | , | ; | : | @ )"""

# qchar        = unreserved / pct-encoded / some-delims

qchar = r"(?: %(unreserved)s | %(pct_encoded)s | %(some_delims)s )" % locals()

# dtext-no-obs = %d33-90 / ; Printable US-ASCII
#              %d94-126  ; characters not including
#                        ; "[", "]", or "\"

dtext_no_obs = r"(?: [\x21-\x5B\x5E-\x7E] )"

# atext           =   ALPHA / DIGIT /    ; Printable US-ASCII
#                     "!" / "#" /        ;  characters not including
#                     "$" / "%" /        ;  specials.  Used for atoms.
#                     "&" / "'" /
#                     "*" / "+" /
#                     "-" / "/" /
#                     "=" / "?" /
#                     "^" / "_" /
#                     "`" / "{" /
#                     "|" / "}" /
#                     "~"
#
# dot-atom-text   =   1*atext *("." 1*atext)

rfc5322_atext = r"""(?: 
    %(ALPHA)s | %(DIGIT)s | 
    ! | # | 
    \$ | %% | 
    & | ' | 
    \* | \+ | 
    - | / | 
    = | \? | 
    \^ | _ | 
    ` | { | 
    \| | } | 
    ~ )""" % locals()
rfc5322_dot_atom_text = r"(?: %(rfc5322_atext)s{1,} (?: . %(rfc5322_atext)s{1,} )* )" % locals()

# FWS             =   ([*WSP CRLF] 1*WSP) /  obs-FWS
#                                        ; Folding white space
#
# ctext           =   %d33-39 /          ; Printable US-ASCII
#                     %d42-91 /          ;  characters not including
#                     %d93-126 /         ;  "(", ")", or "\"
#                     obs-ctext
#
# ccontent        =   ctext / quoted-pair / comment
#
# comment         =   "(" *([FWS] ccontent) [FWS] ")"
#
# CFWS            =   (1*([FWS] comment) [FWS]) / FWS

qcontent = rfc5322_FWS = rfc5322_CFWS = r"(?: )"  ## FIXME

# quoted-string   =   [CFWS]
#                     DQUOTE *([FWS] qcontent) [FWS] DQUOTE
#                     [CFWS]

rfc5322_quoted_string = r"""(?:
    (?: %(rfc5322_CFWS)s )? 
    %(DQUOTE)s (?: (?: %(rfc5322_FWS)s )? %(qcontent)s )* (?: %(rfc5322_FWS)s )? %(DQUOTE)s 
    (?: %(rfc5322_CFWS)s )? )
""" % locals()

# domain       = dot-atom-text / "[" *dtext-no-obs "]"

domain = r"(?: %(rfc5322_dot_atom_text)s | (?: \[ %(dtext_no_obs)s* \] ) )" % locals()

# local-part   = dot-atom-text / quoted-string

local_part = r"(?: %(rfc5322_dot_atom_text)s | %(rfc5322_quoted_string)s )" % locals()

# addr-spec    = local-part "@" domain

addr_spec = r"(?: %(local_part)s @ %(domain)s )" % locals()

# hfvalue      = *qchar

hfvalue = r"(?: %(qchar)s* )" % locals()

# hfname       = *qchar

hfname = r"(?: %(qchar)s* )" % locals()

# hfield       = hfname "=" hfvalue

hfield = r"(?: %(hfname)s = %(hfvalue)s )" % locals()

# to           = addr-spec *("," addr-spec )

to = r"(?: %(addr_spec)s (?: , %(addr_spec)s )* )" % locals()

# hfields      = "?" hfield *( "&" hfield )

hfields = r"(?: \? %(hfield)s (?: & %(hfield)s )* )" % locals()

# mailtoURI    = "mailto:" [ to ] [ hfields ]

mailto_URI = r"(?: mailto : (?: %(to)s )? (?: %(hfields)s )? )" % locals()

### data - RFC2397 (+ RFC2045)

# ietf-token := <An extension token defined by a
#                standards-track RFC and registered
#                with IANA.>

rfc2045_token = r"(?: [\x30-\x7A]+ )"  # FIXME
rfc2045_ietf_token = rfc2045_token
rfc2045_iana_tokens = rfc2045_token

# x-token := <The two characters "X-" or "x-" followed, with
#             no intervening white space, by any token>

rfc2045_x_token = r"(?: [xX] - %(rfc2045_token)s )" % locals()

# extension-token := ietf-token / x-token

rfc2045_extension_token = r"(?: %(rfc2045_ietf_token)s | %(rfc2045_x_token)s )" % locals()

# discrete-type := "text" / "image" / "audio" / "video" /
#                  "application" / extension-token

rfc2045_discrete_type = r"(?: text | image | audio | video | application | %(rfc2045_extension_token)s )" % locals()

# composite-type := "message" / "multipart" / extension-token

rfc2045_composite_type = r"(?: message | multipart | %(rfc2045_extension_token)s )" % locals()

# type := discrete-type / composite-type

rfc2045_type = r"(?: %(rfc2045_discrete_type)s | %(rfc2045_composite_type)s )" % locals()

# subtype := extension-token / iana-token

rfc2045_subtype = r"(?: %(rfc2045_extension_token)s | %(rfc2045_iana_tokens)s )" % locals()

# parameter  := attribute "=" value
# attribute := token
#              ; Matching of attributes
#              ; is ALWAYS case-insensitive.
#
# value := token / quoted-string

rfc2045_quoted_string = r"(?: )"  # FIXME
rfc2045_attribute = r"(?: %(rfc2045_token)s )" % locals()
rfc2045_value = r"(?: %(rfc2045_token)s | %(rfc2045_quoted_string)s )" % locals()
rfc2045_parameter = r"(?: %(rfc2045_attribute)s = %(rfc2045_value)s )" % locals()

# mediatype  := [ type "/" subtype ] *( ";" parameter )

mediatype = r"""(?: 
                    (?: %(rfc2045_type)s / %(rfc2045_subtype)s )? 
                    (?: ; %(rfc2045_parameter)s )*
                )""" % locals()

# uric          = reserved | unreserved | escaped  // 2396
# data       := *urlchar

rfc2396_uric = r"(?: %(reserved)s | %(unreserved)s | %(pct_encoded)s )" % locals()
data = r"(?: %(rfc2396_uric)s* )" % locals()

# dataurl    := "data:" [ mediatype ] [ ";base64" ] "," data

data_URI = r"(?: data : (?: %(mediatype)s )? (?: ;base64 )? , %(data)s )" % locals()

### gopher - RFC4266

# gopher://<host>:<port>/<gopher-path>

gopher_path = path
gopher_URI = r"(?: gopher :// %(host)s : %(port)s / %(gopher_path)s )" % locals()

### file - draft-kerwin-file-scheme-13

# f-scheme       = "file"

file_f_scheme = r"(?: file )"

# f-auth         = [ userinfo "@" ] host

file_f_auth = r"(?: (?: %(userinfo)s @ )? %(host)s )" % locals()

# unc-path       = 2*3"/" authority path-absolute

file_unc_path = r"(?: /{2,3} %(authority)s %(path_absolute)s )" % locals()

# drive-marker   = ":" / "|"

file_drive_marker = r"(?: : | \| )"

# drive-letter   = ALPHA [ drive-marker ]

file_drive_letter = r"(?: %(ALPHA)s (?: %(file_drive_marker)s )? )" % locals()

# windows-path   = drive-letter path-absolute

file_windows_path = r"(?: %(file_drive_letter)s %(path_absolute)s )" % locals()

# local-path     = path-absolute
#                / windows-path

file_local_path = r"(?: %(path_absolute)s | %(file_windows_path)s )" % locals()

# auth-path      = [ f-auth ] path-absolute
#                / unc-path
#                / windows-path

file_auth_path = r"""(?: 
                          (?: %(file_f_auth)s? %(path_absolute)s ) 
                          | %(file_unc_path)s 
                          | %(file_windows_path)s 
                       )""" % locals()

# f-hier-part    = "//" auth-path
#                / local-path

file_f_hier_part = r"(?: (?: // %(file_auth_path)s ) | %(file_local_path)s )" % locals()

# file-URI       = f-scheme ":" f-hier-part [ "?" query ]

file_URI = r"(?: %(file_f_scheme)s : %(file_f_hier_part)s (?: \? %(query)s )? )" % locals()

if "__main__" == __name__:
    import re
    import sys

    try:
        instr = sys.argv[1]
    except IndexError:
        print("usage: %s test-string" % sys.argv[0])
        sys.exit(1)

    print('testing: "%s"' % instr)

    print("URI:", end=' ')
    if re.match("^%s$" % URI, instr, re.VERBOSE):
        print("yes")
    else:
        print("no")

    print("URI reference:", end=' ')
    if re.match("^%s$" % URI_reference, instr, re.VERBOSE):
        print("yes")
    else:
        print("no")

    print("Absolute URI:", end=' ')
    if re.match("^%s$" % absolute_URI, instr, re.VERBOSE):
        print("yes")
    else:
        print("no")

    scheme = instr.split(":", 1)[0].lower()
    scheme_validator = locals().get("%s_URI" % scheme, None)
    if scheme_validator:
        print("'%s' URI: " % scheme, end=' ')
        if re.match("^%s$" % scheme_validator, instr, re.VERBOSE):
            print("yes")
        else:
            print("no")



# Copyright (c) 2016-2021 Damon Lynch
# SPDX - License - Identifier: MIT

from enum import Enum


class FileManagerType(Enum):
    regular = 1  # file_manager "File1" "File2"
    select = 2  # file_manager --select
    dir_only_uri = 3  # cannot select files
    show_item = 4  # file_manager --show-item
    show_items = 5  # file_manager --show-items
    win_select = 6  # explorer.exe /select
    reveal = 7  # open --reveal   (macOS)
    dual_panel = 8  # file_manager "File1" "File2"


class Platform(Enum):
    windows = 1
    linux = 2
    macos = 3


single_file_only = ("explorer.exe", "pcmanfm", "open", "cutefish-filemanager")
cannot_open_uris = ("fman", "fman.exe", "lumina-fm")



# Copyright (c) 2021-2024 Damon Lynch
# Some portions Copyright (c) 2008-2021 The pip developers
# SPDX - License - Identifier: MIT

import os
import re
import shlex
from collections import defaultdict
from pathlib import Path
from typing import DefaultDict, List
from urllib.parse import urljoin, urlparse
from urllib.request import pathname2url, url2pathname

from ..constants import Platform, cannot_open_uris
from . import current_platform, urivalidate


def filemanager_requires_path(file_manager: str) -> bool:
    return current_platform == Platform.windows or file_manager in cannot_open_uris


def is_uri(path_uri: str) -> bool:
    """
    Checks if string is probably a uri of some kind.
    :param path_uri: the
    :return: True if probably a URI, else False
    """

    if path_uri and path_uri.startswith("camera:/"):
        return True
    return re.match("^%s$" % urivalidate.URI, path_uri, re.VERBOSE) is not None


def quote_path(path: Path) -> Path:
    """
    Quote path in a way that works with file managers on Windows and Unix-like.

    If path is already quoted, returns it as is.

    Replaces single quoted string with double quotes on Windows.

    Uses shlex.quote on non-Windows platforms, but again only if the path is
    not already quoted.

    :param path: path to quote, if necessary
    :param target_platform: platform the file manager command will be executed on
    :return: double-quoted path
    """

    p = str(path)
    if current_platform == Platform.windows:
        # Double quotes are not allowed in paths names - they are used for quoting

        if re.match("""'(.*)'""", p) is not None:
            # Replace single quotes with double quotes
            return Path(f'"{p[1:-1]}"')
        if re.match(r"""\"(.*)\"""", p) is None:
            # Add double quotes where there was no quoting at all
            return Path(f'"{path}"')
    else:
        if not (p[0] in ('"', "'") and p[-1] == p[0]):
            return Path(shlex.quote(p))
    return path


def path_to_file_uri(path: str) -> str:
    """
    Convert a path to a file: URL.  The path will be made absolute and have
    quoted path parts.

    Taken from pip: https://github.com/pypa/pip/blob/main/src/pip/_internal/utils/urls.py
    Copyright (c) 2008-2021 The pip developers
    """

    path = os.path.normpath(os.path.abspath(path))
    url = urljoin("file:", pathname2url(path))
    return url


def file_uri_to_path(uri: str) -> str:
    """
    Convert a file: URL to a path.

    On Windows this is more reliable than urllib.parse.urlparse, because that fails when
    run with a URI like file:///D:/some/directory

    Taken from https://stackoverflow.com/a/61922504/592623
    and modified by Damon Lynch 2021, 2024
    """

    parsed = urlparse(uri)
    host = f"{os.path.sep}{os.path.sep}{parsed.netloc}{os.path.sep}"
    p = os.path.normpath(os.path.join(host, url2pathname(parsed.path)))
    return p


def directories_and_their_files(paths: List[str]) -> DefaultDict[str, List[str]]:
    """
    Group paths into directories and their files.

    If path is a directory, the parent will be the directory, and the subfolder will
    be the child of that directory.

    :param paths: list of paths
    :return: default dict of folders with list of their files
    """

    if isinstance(paths, str):
        paths = [paths]
    folder_contents = defaultdict(list)
    for path in paths:
        p = Path(path)
        folder_contents[str(p.parent)].append(p.name)
    return folder_contents
